/******************************************************************************/
/*!
\file		MultiThread.cpp
\project	Icarus Engine
\author		Jorge Antonio Jimenez

\brief		Generates multiple Threads depending on the Cores on machine.
			To use it to paralize work.
*/
/******************************************************************************/
#include <thread>
#include <algorithm>

#include <Engine\MultiThread\MultiThread.h>
#include <Engine\DebugDiagnostics.h>

namespace IcarusEngine
{
	MultiThread* MULTITHREAD = nullptr;
	
	unsigned int minimum_threads = 2;

	void DoJob(JobPtr job)
	{
		if(job)
		{
			// Specifies that the job is being done.
			job->SetStatus(JobStatus::Running);
			// Doing the Job.
			job->Execute();
			// Specify that the job is done, so the owner can work with its data.
			job->SetStatus(JobStatus::Completed);
		}
	}

	/******************************************************************************/
	/*!
	  \brief
		System in charge of working with concurrency jobs.
	*/
	/******************************************************************************/
	MultiThread::MultiThread()
		: m_Working(false),
		m_NotWorking(0)
	{
		ErrorIf(MULTITHREAD, "MultiThread System has already been created.");
		MULTITHREAD = this;
	}
	
	/******************************************************************************/
	/*!
	  \brief
		Shutdown the remaining open threads
	*/
	/******************************************************************************/
	MultiThread::~MultiThread()
	{
		Shutdown();
	}

	/******************************************************************************/
	/*!
	  \brief
		Handles different messages relevant to the MultiThreaded system
	
	   \return
		void
	*/
	/******************************************************************************/
	void MultiThread::HandleMessage(Message* message)
	{
		// Review the Messages.
		switch (message->MessageId)
		{
			case MessageType::DecreaseThreads:
			{
				m_NotWorking = std::min(m_Workers.size()-static_cast<int>(minimum_threads), static_cast<unsigned int>(++m_NotWorking));
				for(unsigned int i = 0; i < m_Workers.size(); ++i)
				{
					m_Workers[i].SetStatus(i < (m_Workers.size()-m_NotWorking));
				}
				return;
			}
			case  MessageType::IncrementThreads:
			{
				m_NotWorking = std::max(--m_NotWorking, 0);
				for(unsigned int i = 0; i < m_Workers.size(); ++i)
				{
					m_Workers[i].SetStatus(i < (m_Workers.size()-m_NotWorking));
				}
				return;
			}
			case MessageType::SwitchMultithread:
			{
				m_Working = ((m_Workers.size() != 0) ? !m_Working : false);
				return;
			}
		}
	}
	
	/******************************************************************************/
	/*!
	  \brief
		Starts up the Multi-Threaded system
	
	   \return
		void
	*/
	/******************************************************************************/
	void MultiThread::Initialize()
	{
		// Review how many concurrent jobs the hardware is capable of.
		unsigned int size = std::thread::hardware_concurrency();

		// If it can't reach a minimum thread amount it won't start multithreading
		// for safety reasons.
		if(size < minimum_threads)
		{
			return;
		}
		m_Working = true;
		PrintOut("Opening %i Threads", size);
		// Creates that amount of workers
		m_Workers.resize(size);

		for(unsigned int i=0; i<size; ++i)
		{
			// Start the workers to do jobs.
			m_Workers[i].StartUp(i+1);
		}
	}

	/******************************************************************************/
	/*!
	  \brief
		Keeps track and updates the system dependind on the jobs.
	
	   \return
		void
	*/
	/******************************************************************************/
	void MultiThread::Update(float dt)
	{
		if(!m_Tracker.empty())
		{
			while(!m_Tracker.empty())
			{
				JobPtr job = nullptr;
				m_Tracker.try_pop(job);
				if(job)
				{
					m_JobTracker.push_back(job);
				}
			}
		}
		// Review if there is any work on the stack.
		if(!m_JobTracker.empty())
		{
			// Work iterator.
			auto it = m_JobTracker.begin();
			while(it != m_JobTracker.end())
			{
				JobPtr job = (*it);
				// Increase time of all the jobs in the queue
				job->AddAge(dt);
				// If the job is still on the queue and
				// a more than reasonable amount of time has pass
				if(job->TimeOut() && JobIsWaiting(job))
				{
					//Upgrade priority and reset the timer.
					job->PriorityUpgrade();
					job->TimeReset();
					// Reinsert the job with the new priority
					UpdatingJob(job);
				}
				// If the job is no longer waiting remove from the tracking.
				if(!JobIsWaiting(job))
				{
					auto temp = it++;
					m_JobTracker.erase(temp);
				}
				else
				{
					++it;
				}
			}
		}

		// Prints how many jobs are pending on the Queue.
		PrintOut("Pending Jobs %i", m_JobTracker.size());
	}

	/******************************************************************************/
	/*!
	  \brief
		Returns the name of this system
	
	   \return
		std::string - The name of the System
	*/
	/******************************************************************************/
	std::string MultiThread::GetName()
	{
		return "MultiThread";
	}

	/******************************************************************************/
	/*!
	  \brief
		Add a job to the system and kepts track of given job.
	
	   \return
		void
	*/
	/******************************************************************************/
	void MultiThread::AddingJob(JobPtr job)
	{
		if(m_Working)
		{
			m_Tracker.push(job);
			UpdatingJob(job);
		}
		else
		{
			DoJob(job);
		}
	}

	/******************************************************************************/
	/*!
	  \brief
		Add a job to the job queue, depending on its priority.
	
	   \return
		void
	*/
	/******************************************************************************/
	void MultiThread::UpdatingJob(JobPtr job)
	{
		m_PendingJobs[static_cast<unsigned int>(job->GetPriority())].push(job);
	}

	/******************************************************************************/
	/*!
	  \brief
		Thread safe, it removes a job from the queue and give it to the one that
		requested it. In case that the job has already been done at that point 
		it will remove it and kept looking.
	
	   \return
		JobPtr - a shared pointer to an assign job, if is null there is no work for now.
	*/
	/******************************************************************************/
	JobPtr MultiThread::GetNextJob()
	{
		for(unsigned int i=0; i<static_cast<unsigned int>(JobPriority::MaxPriority); ++i)
		{
			auto index = static_cast<unsigned int>(JobPriority::High)-i;
			if(!m_PendingJobs[index].empty())
			{
				// Review the job list until is empty.
				while(!m_PendingJobs[index].empty())
				{
					JobPtr job = nullptr;
					// Try to remove a job if a job still exist.
					// Since is multi-threaded there is the change of failure given
					// that another thread might have take it already.
					m_PendingJobs[index].try_pop(job);
					// If there is a job and its on a waiting state
					if(job && JobIsWaiting(job))
					{
						// Give that job to someone.
						return job;
					}
				}
			}
		}

		return nullptr;
	}

	/******************************************************************************/
	/*!
	  \brief
		Makes sure that all the workers stop working before closing down the system.
	
	   \return
		void
	*/
	/******************************************************************************/
	void MultiThread::Shutdown()
	{
		if(m_Workers.empty())
		{
			return;
		}
		for(unsigned int i=0; i < m_Workers.size(); ++i)
		{
			m_Workers[i].ShutDown();
		}
	}

	/******************************************************************************/
	/*!
	  \brief
		Review if a job is on a waiting state or not.
	
	   \return
		void
	*/
	/******************************************************************************/
	bool MultiThread::JobIsWaiting(JobPtr job)
	{
		return job->GetStatus() == JobStatus::Waiting;
	}

	bool MultiThread::isWorking()
	{
		return m_Working;
	}
}