/******************************************************************************/
/*!
\file		Worker.cpp
\project	Icarus Engine
\author		Jorge Antonio Jimenez

\brief		Generates a single Thread, and keeps track of the activity
			that the thread is executing.
*/
/******************************************************************************/

#include <Engine\Core\Core.h>
#include <Engine\DebugDiagnostics.h>
#include <Engine\MultiThread\Worker.h>
#include <Engine\MultiThread\MultiThread.h>

namespace IcarusEngine
{
	
	/******************************************************************************/
	/*!
	  \brief
		Get next job in line and executes it
	
	   \return
		void
	*/
	/******************************************************************************/
	void GetandExecuteJob(Worker * worker)
	{
		if(!MULTITHREAD)
		{
			PrintOut("I'm not suppose to be here.");
			return;
		}
		// Gets the next job in line.
		JobPtr job = MULTITHREAD->GetNextJob();
		if(job)
		{
			if(worker)
			{
				//PrintOut("Worker %i is working in %s.", worker->GetWorkerId(), job->GetName().c_str());
				// Specify what job does the thread is working on.
				worker->SetJob(job);
			}
			else
			{
				//PrintOut("Main Thread is working in %s.", job->GetName().c_str());
			}
			DoJob(job);
		}
		else
		{
			if(MULTITHREAD->isWorking())
			{
				PrintOut("There are no Pending Jobs.");
			}
		}
		if(worker)
		{
			// Set up the worker as null, to specify that it don't have work.
			worker->SetJob(nullptr);
		}
	}

	/******************************************************************************/
	/*!
	  \brief
		Function that will be executed in different threads
	
	   \return
		void
	*/
	/******************************************************************************/
	void DoingWork(Worker* worker)
	{
		// While the game is running, do the following.
		while(CORE->IsActive())
		{
			// Look if the worker is active or not. This is for testing purposes.
			if(worker->IsActive())
			{
				GetandExecuteJob(worker);
				//Makes sure computer doesn't go kablooey.
				std::this_thread::sleep_for(std::chrono::milliseconds(1));
			}
			else
			{
				// Sleep longer because the worker is not active for debugging purpuses.
				//PrintOut("Worker %i is slacking off again.", worker->GetWorkerId());
				std::this_thread::sleep_for(std::chrono::milliseconds(1000));
			}
		}
		PrintOut("Worker %i has quit.", worker->GetWorkerId());
	}

	/******************************************************************************/
	/*!
	  \brief
		A Worker that will execute a job.
	*/
	/******************************************************************************/
	Worker::Worker()
		: m_Active(true),
		m_WorkerId(0),
		m_CurrentJob(nullptr),
		m_workerthread(nullptr)
	{
	}

	/******************************************************************************/
	/*!
	  \brief
		Makes sure to finish and close the thread.
	*/
	/******************************************************************************/
	Worker::~Worker()
	{
		ShutDown();
	}

	/******************************************************************************/
	/*!
	  \brief
		Give the worker an Id and Generates a new thread, to work with.
	
	   \return
		void
	*/
	/******************************************************************************/
	void Worker::StartUp(unsigned int index)
	{
		m_WorkerId = index;
		m_workerthread = new std::thread(DoingWork, this);
	}

	/******************************************************************************/
	/*!
	  \brief
		Makes sure to clean all the data.
	
	   \return
		void
	*/
	/******************************************************************************/
	void Worker::ShutDown()
	{
		// Checks if the thread still exist.
		if(m_workerthread)
		{
			// Merge it with the main thread.
			m_workerthread->join();
			// When finish working it delete it.
			SafeDelete(m_workerthread);
		}
	}

	/******************************************************************************/
	/*!
	  \brief
		Tells the Worker what is his current job.
	
	  \param
		JobPtr - a shared pointer to the job that the worker is being assign.

	   \return
		void
	*/
	/******************************************************************************/
	void Worker::SetJob(JobPtr job)
	{
		m_CurrentJob = job;
	}

	/******************************************************************************/
	/*!
	  \brief
		Returns the current job of the Worker.
	
	   \return
		JobPtr - a shared pointer to the job that its being executed.
	*/
	/******************************************************************************/
	JobPtr Worker::GetJob()
	{
		return m_CurrentJob;
	}

	/******************************************************************************/
	/*!
	  \brief
		Review if the worker is being active to work, or it has been unactivated,
		to simulate slow down.
	
	   \return
		bool - The current status of the worker.
	*/
	/******************************************************************************/
	bool Worker::IsActive()
	{
		return m_Active;
	}

	/******************************************************************************/
	/*!
	  \brief
		Sets up the worker status in order to work or not.
	
	   \return
		void
	*/
	/******************************************************************************/
	void Worker::SetStatus(bool status)
	{
		m_Active = status;
	}

	/******************************************************************************/
	/*!
	  \brief
		Return the id of the worker so it can be identify.
	
	   \return
		void
	*/
	/******************************************************************************/
	unsigned int Worker::GetWorkerId()
	{
		return m_WorkerId;
	}
}